"""
======================================================
Проект: Универсальный конфигуратор
Модуль: Хранение конфигурации
Автор:  Андрей Орлов
Версия: ${Id: }
------------------------------------------------------
Дата:   Mon Nov 19 21:09:12 MSK 2001
======================================================
"""

import UserDict
import sys
from cPickle import dump,load
from Exceptions import *

import Parser

class BaseConfig (UserDict.UserDict) :
        """ Универсальный конфигуратор / хранение конфигурации : Базовый
        класс -- конфигурация сохраняется посредством последовательности
        словарей, в которых каждый последующий наследует атрибуты
        предыдущего. """

        def __init__(self,dict=None, parent={},realm='',arguments=[]) :
                self.parent = parent            # Родительский объект
                self.realm = realm              # Название данной секции
                self.arguments = arguments      # Ссылка на массив аргументов командной строки
                self.sections = []              # Секции
                UserDict.UserDict.__init__(self,dict)

        def __len__(self): return len(self.data) + self.parent.__len__()

        def __getattr__(self,attr) :
                if attr == 'application' :
                        try :
                                return UserDict.UserDict.__getattr__(self,attr)
                        except AttributeError :
                                self.application = getattr(self.parent,attr)
                                return self.application
                else :
                        return UserDict.UserDict.__getattr__(self,attr)

        def __getitem__(self,key) :
                try :
                        return self.data[key]
                except KeyError :
                        return self.parent[key]

        def __setitem__(self,key,value) :
                s = 'switch_' + key
                if hasattr(self,s) and callable(getattr(self,s)) :
                        apply(getattr(self,s),(key,value),{})
                else :
                        self.data[key] = value

        def use(self,key) :
                s = 'switch_' + key
                if hasattr(self,s) and callable(getattr(self,s)) :
                        try :
                                apply(getattr(self,s),(key,),{})
                        except TypeError :
                                raise ConfigError,'Ключ %s требует аргумент' % key
                else :
                        self[key] = self.default(key)

        def setparent(self,cfg) :       parent = self.cfg

        def keys(self) :        return self.data.keys() + self.parent.keys()
        def items(self) :       return self.data.items() + self.parent.items()
        def values(self) :      return self.data.values() + self.parent.values()
        def has_key(self, key) : return self.data.has_key(key) or self.parent.has_key(key)
     

        def duplicate(self,dict=None,realm='') :
                return self.__class__(parent=self,realm=realm,dict=dict,arguments=self.arguments)

        def default(self,attr) :
                return 1

        def section(self,dict=None,realm='') :
                try :
                        if dict :
                                self.data[realm].update(dict)
                        return self.data[realm]
                except KeyError :
                        self.sections.append(realm)
                        return self.__class__(parent=self,realm=realm,dict=dict,arguments=self.arguments)

        def parents(self) :
                ps = []
                cr = self               
                while type(cr) is not type({}) :
                        ps.append(cr)
                        cr = cr.parent
                ps.append(cr)
                return ps

        def prnt(self,min=0,max=-1) :
                for item in self.parents()[min:max] :
                        print '===> %-20s =' % item.realm + '=' * 45
                        self.save(cfg=item.data,plain=1)
                print '=' * 72

        def getdict(self) :
                dict = self.parent.getdict()            
                dict.update(self.dict)
                return dict

        def dictionary(self,min=0,max=-1) :
                ps = self.parents()
                cfg = {}
                for d in ps[min:max] :
                        cfg.update(d.data)
                return cfg

        def filtered(self,*items) :
                cfg = {}
                for item in items :
                        try :
                                cfg[item] = self[item] 
                        except KeyError :
                                pass
#               print cfg
                return cfg

        def save(self,ouf=sys.stdout,cfg=None,plain=1) :
                if cfg is None : 
                        cfg = self.data

                ks = cfg.keys()
                ks.sort()

                for item in filter(lambda x,cfg=cfg : not isinstance(cfg[x],BaseConfig), ks ) :
                        if type(cfg[item]) is type('') :
                                s = '"%s"' % cfg[item]
                        else :  
                                s = str(cfg[item])
                        ouf.write( "%-40s%s\n" % ("%s = " % item,str(s)) )

                for item in filter(lambda x,cfg=cfg : isinstance(cfg[x],BaseConfig), ks ) :
                        print "\n[%s]\n" % item
                        if plain :
                                cfg[item].save(cfg=cfg[item].data,plain=plain)
                        else :
                                cfg[item].save(plain=plain)


class Switch :
        """\
Описание:       Универсальный конфигуратор / хранение конфигурации :
                Предопределенные ключи. Обработка ключей, требуемая для конфигурации
                Application.
Автор:          Андрей Орлов
Дата:           Wed Nov 21 13:09:53 MSK 2001
Ключи:          
                verbose         - увеличить (задать) уровень трассировки,
                verbose-stderr  - выводить трассировку на stderr,
                debug           - запуск в отладочном режиме,
                
                dump = type     - вывести дамп конфигурации, тип может быть:
                        detail  - детальная схема конфигурации,
                        content - список секций конфигурации,
                        plain   - конфигурационный файл
                dump-min        - минимальный уровень дампа конфигурации,
                dump-max        - максимальный уровень дампа конфигурации,

                help = [chapter]- вывести хелп,         
                exit            - прекратить исполнение программы и выйти,
                
                cfg_dump        - сохранить конфигурацию в виде бинарного дампа,
                cfg_load        - загрузить конфигурацию из бинарного дамп,
                cfg_read        - прочитать конфигурационный файл,
                cfg_clean       - восстановить умолчания в конфигурации.
"""
        
        default_cfg = {
                'verbose-level' : 1,
                'verbose-stderr': 0,
                'debug'         : 0,
                'dump-level'    : 0,
                'dump-min'      : 2,
                'dump-max'      : -2,
        }

        def __init__(self) :
                if type(self.parent) is type({}):
                        for attr,value in self.default_cfg.items() :    
                                if not self.data.has_key(attr) :
                                        self.data[attr] = value

        def switch_verbose(self,attr,value=1) :
                if value is not None :
                        self['verbose-level'] = int(value)
                else :
                        self['verbose-level'] = self['verbose-level'] + 1

        def switch_dump(self,attr,value='detail') :
                if value == 'detail' :
                        print '# Детальный дамп конфигурации [%i:%i]' % (self['dump-min'],self['dump-max'])
                        self.prnt(self['dump-min'],self['dump-max'])

                elif value == 'plain' :
                        print '# Плоский дамп конфигурации [%i:%i]' % (self['dump-min'],self['dump-max'])
                        self.save(cfg=self.dictionary(self['dump-min'],self['dump-max']))

                elif value == 'content' :
                        print '# Очередь конфигураций [%i:%i]' % (self['dump-min'],self['dump-max'])
                        n = self['dump-min']
                        for item in self.parents()[self['dump-min']:self['dump-max']] :                 
                                print n,":",item.realm
                                n = n+1
                else :
                        print '# Возможный тип дампа - detail,plain,content'

        def switch_help(self,attr,value=None) :
                if value == 'cfg' :
                        print self.__doc__
                else :
                        cfg = {'application':sys.argv[0], 'prefix':sys.prefix}
                        cfg.update(self.dictionary(0,-1))
                        print self.application.__doc__ % cfg

        def switch_exit(self,attr,value=None) :
                sys.exit(0)

        def switch_cfg_dump(self,attr,value) :
                try :
                        f = open(value,'w')
                        dump(self,f)
                        f.close()
                except IOError,msg :
                        raise ConfigError,'Не удается сохранить дамп конфигурации в (%s): %s' % (value,msg)

        def switch_cfg_load(self,attr,value) :
                """ Загрузка конфигурации из ранее сохраненного дампа """
                try :
                        value = str(value)
                        if not hasattr(self,'hasbeenloaded') :
                                f = open(value,'r')
                                self.parent = load(f)
                                f.close()
                                self.hasbeenloaded = 1
                except IOError,msg :
                        raise ConfigError,'Не удается загрузить дамп конфигурации (%s): %s' % (value,msg)
                except EOFError,msg :
                        raise ConfigError,'Не удается загрузить дамп конфигурации (%s): внезапный конец файла' % (value,)

        def switch_cfg_read(self,attr,value) :
                self.parent = self.__class__(realm=value,parent=self.parent)
                Parser.ParseFile(self.parent,open(value,'r'))

        def switch_cfg_clean(self,attr,value=None) :
                if value :
                        try :
                                self[value] = self.parents()[-2][value]
                        except KeyError :
                                raise ConfigError,'Неизвестный параметр "%s"' % value
                else :
                        self.parent = self.parents()[-2]

class Strict :
        """ Универсальный конфигуратор / хранение конфигурации : Строгая
        загрузка конфига - допускает загрузку переменных только из числа уже
        определенных, причем грузятся переменные того же типа."""

        def __setitem__(self,key,value) :
                s = 'switch_' + key
                if hasattr(self,s) and callable(getattr(self,s)) :
                        apply(getattr(self,s),(key,value),{})
                else :
                        if not isinstance(value,BaseConfig) :
                                if not self.has_key(key) :
                                        raise ConfigError,'Неизвестный ключ "%s"' % key
                                elif type(self[key]) != type(value) :
                                        raise ConfigError,'Неверный тип данных ключа "%s" (%s -> %s)' % (key,type(value),type(self[key]))
                        self.data[key] = value
                
class Config (Switch, BaseConfig) :
        __doc__ = """\
Описание:       Универсальный конфигуратор / хранение конфигурации : Простая
                конфигурация.  
""" + Switch.__doc__

        def __init__(self,dict=None, parent={},realm='',arguments=[],cfg_load=None) :
                if cfg_load :
                        self.cfg_load('cfg_load',cfg_load)
                else :
                        BaseConfig.__init__(self,dict,parent,realm,arguments)
                        Switch.__init__(self)


class ConfigS (Strict, Config) :
        __doc__ = """\
Описание:       Универсальный конфигуратор / хранение конфигурации : Простая
                конфигурация с проверкой.
""" + Switch.__doc__



        
        
